/* Elvira Bluetooth FTP 
 * Copyright (C) 2007  Alberto Salas Bragado
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "bftp_client.h"

static obexftp_client_t *cli= NULL;

int bftp_hci_search(device_t **list);

int bftp_search(device_t **list) {

	int numdisp;
	device_t *lista;
	
	numdisp = bftp_hci_search(&lista);
	*list = lista;
	
	return numdisp;
	
}

int bftp_connect(device_t *device) {

	int ret;

	//Abrimos conexión OBEX Bluetooth
	cli = obexftp_open(OBEX_TRANS_BLUETOOTH, NULL, NULL, NULL);
	if (cli == NULL)
		return E_FTP_LOCAL_CONN;

	//Conectamos con el servidor OBEX FTP
	ret = obexftp_connect(cli, device->mac, device->port); /*!!!*/
	if (ret < 0) {
		obexftp_close(cli);
		cli = NULL;
		return E_FTP_REMOTE_CONN;
	}

	//Conexión sin errores
	return 0;

}

int bftp_disconnect() {

	if (cli != NULL) {
	
		int ret;

		//Desconectamos del servidor OBEX FTP
		ret = obexftp_disconnect(cli);

		//Cerramos la conexion OBEX Bluetooth
		obexftp_close(cli);
		cli = NULL;

		//Devolvemos el resultado de la operación
		if (ret < 0)
			return E_FTP_DISCONNECTING;
	}
	
	return 0;
	
}

int bftp_state() {
	
	if (cli == NULL)
		
		return FTP_DISCONNECTED;
		
	else
		
		return FTP_CONNECTED;
	
}

int bftp_list(char *directory, file_t **list) {

	int tam=0;
	file_t *lista= NULL;
	//Variables de libxml2 necesarias para parsear el documento XML
	xmlDocPtr doc;
	xmlNodePtr cur;

	//Obtenemos el listado de los ficheros del directorio en XML
	if (obexftp_list(cli, NULL, directory) < 0) {
		return E_FTP_LIST;
	}

	//Creamos el objeto Document de libxml2
	doc = xmlReadMemory((const char *) cli->buf_data, (int) cli->buf_size,
			NULL, NULL, 0);
	//Movemos el cursor al principio del documento
	cur = xmlDocGetRootElement(doc);

	//Comprobamos si el documento esta vacío
	if (cur == NULL) {
		fprintf(stderr,"Documento Vacío\n");
		xmlFreeDoc(doc);
		return E_FTP_LIST_EMPTY;
	}

	//Comprobamos que el nodo raiz sea el correcto
	if (xmlStrcmp(cur->name, (const xmlChar *) "folder-listing")) {
		fprintf(stderr,"Documento de tipo incorrecto, root node != folder-listing\n");
		xmlFreeDoc(doc);
		return E_FTP_LIST_CORRUPT;
	}

	//Movemos el cursor al primer hijo
	cur = cur->xmlChildrenNode;

	//Repetimos hasta que no haya mas hijos
	while (cur != NULL) {

		//Comprobamos que el nodo actual sea un folder
		if ((!xmlStrcmp(cur->name, (const xmlChar *) "folder"))) {

			//Incrementamos el tamaño de la lista a devolver
			tam++;
			//Ampliamos el tamaño de la memoria dinámica solicitada para añadir el folder actual
			lista = (file_t *) realloc(lista, sizeof(file_t)*tam);

			//Recuperamos las propiedades del nodo folder actual
			lista[tam-1].type = FILET_FOLDER;
			sprintf(lista[tam-1].name, "%s", xmlGetProp(cur,
					(const xmlChar *) "name"));
			if (xmlGetProp(cur, (const xmlChar *) "created") != NULL) {
				sprintf(lista[tam-1].created, "%s", xmlGetProp(cur,
						(const xmlChar *) "created"));
			}
			sprintf(lista[tam-1].user_perm, "%s", xmlGetProp(cur,
					(const xmlChar *) "user-perm"));
			sprintf(lista[tam-1].mem_type, "%s", xmlGetProp(cur,
					(const xmlChar *) "mem-type"));
			strcpy(lista[tam-1].size, "");

			//Si no, comprobamos que el nodo actual sea un file
		} else if ((!xmlStrcmp(cur->name, (const xmlChar *) "file"))) {

			//Incrementamos el tamaño de la lista a devolver
			tam++;
			//Ampliamos el tamaño de la memoria dinámica solicitada para añadir el file actual
			lista = (file_t *) realloc(lista, sizeof(file_t)*tam);

			//Recuperamos las propiedades del nodo file actual
			lista[tam-1].type = FILET_FILE;
			sprintf(lista[tam-1].name, "%s", xmlGetProp(cur,
					(const xmlChar *) "name"));
			sprintf(lista[tam-1].size, "%s", xmlGetProp(cur,
					(const xmlChar *) "size"));
			sprintf(lista[tam-1].modified, "%s", xmlGetProp(cur,
					(const xmlChar *) "modified"));
			sprintf(lista[tam-1].user_perm, "%s", xmlGetProp(cur,
					(const xmlChar *) "user-perm"));

		}

		//Nos movemos al siguiente nodo
		cur = cur->next;

	}

	//Liberamos el objeto Document de libxml2
	xmlFreeDoc(doc);

	//Devolvemos la lista de ficheros creada
	*list = lista;
	//Devolvemos el tamaño de lista de ficheros
	return tam;
}

int bftp_put(char *local_route, char *remote_route) {

	int ret;

	//Mensaje debug
	//printf("\n* Subiendo %s a %s ... OK!\n", local_route, remote_route);

	//Solicitamos el envío de un fichero al servidor y comprobamos la respuesta
	ret = obexftp_put_file(cli, local_route, remote_route);
	if (ret < 0) {
		return E_FTP_PUT;
	}

	//Volvemos al directorio raiz
	//Es una "fallo" de obexftp
	obexftp_cdtop(cli);

	//Devolvemos el resultado de la operacion
	return (ret);
}

int bftp_get(char *local_route, char *remote_route) {

	int ret;

	//Mensaje debug
	//printf("\n* Descargando %s a %s ... OK!\n", remote_route, local_route);

	//Solicitamos la recepción de un fichero desde el servidor
	ret = obexftp_get(cli, local_route, remote_route);
	if (ret < 0) {
		return E_FTP_GET;
	}

	//Volvemos al directorio raiz
	//Es una "fallo" de obexftp
	obexftp_cdtop(cli);

	//Devolvemos el resultado de la operación
	return (0);
}

int bftp_del(char *remote_route) {

	int ret;

	//Solicitamos la eliminación del fichero remoto
	ret = obexftp_del(cli, remote_route);
	if (ret < 0) {
		return E_FTP_DEL;
	}

	//Volvemos al directorio raiz
	//Es una "fallo" de obexftp
	obexftp_cdtop(cli);

	//Devolvemos el resultado de la operación
	return (ret);
}

int bftp_mkdir(char *remote_route) {

	int ret;

	//Solicitamos la creación del directorio en el servidor
	ret = obexftp_mkpath(cli, remote_route);
	if (ret < 0) {
		return E_FTP_DEL;
	}

	//Volvemos al directorio raiz
	//Es una "fallo" de obexftp
	obexftp_cdtop(cli);

	//Devolvemos el resultado de la operación
	return (ret);
}

void bftp_free(void **pointer) {

	//Liberamos la memoria dinámica
	free(*pointer);

	//Puntero a null para evitar confusiones
	*pointer = NULL;

}

/* http://people.csail.mit.edu/albert/bluez-intro/c404.html */
int bftp_hci_search(device_t **list) {
	
	//Puntero a estructura con informacion de un dispositivo, resultado de una busqueda
	inquiry_info *ii= NULL;
	
	//Duración de la búsqueda de dispositivos = len * 1.28 segundos
	int len = 5;
	
	//Maximo numero de dispositivos a devolver por BlueZ y numero de dispositivos encontrados
	int max_rsp = 256, num_rsp;
	
	//Numero de dispositivos devueltos por esta funcion
	//(Sólo los que exporten FTP)
	int num_ftp=0;
	
	//Identificador de dispositivo y socket del microcontrolador
	int dev_id, sock;
	
	//Flags para la busqueda de dispositivos
	//IREQ_CACHE_FLUSH obliga a borrar los dispositivos de la cache de antiguas búsquedas
	//antes de iniciar una nueva.
	int flags = IREQ_CACHE_FLUSH;
	
	//Resto de variables necesarias
	int i;
	device_t *lista = NULL;
	char mac_aux[100];
	int port_aux;

	//A los dispositivos Bluetooth se les asigna un número entero identificativo, empezando desde 0.
	//Para usar un dispositivo es necesario solicitarlo usando su número identificativo.
	//Como sólo suele haber un dispositivo en el sistema, con NULL indicamos que queremos usar el primer
	//dispositivo disponible. (Nota: no es buena idea usar un 0 en esta funcion sin comprobar si esta disponible).
	dev_id = hci_get_route(NULL);
	
	//Solicitamos la conexión con el microcontrolador del dispositivo a usar, NO con un dispositivo remoto.
	//Es necesario para algunas operaciones.
	sock = hci_open_dev(dev_id);
	
	//Comprobamos si hemos obtenenido un dispositivo y socket.
	if (dev_id < 0 || sock < 0) {
		/*perror("Accediendo al dispositivo local");
		exit(1);*/
		return E_LOCAL_DISP;
	}

	//Lanzamos la busqueda de dispositivos que nos devolvera una lista de dispositivos encontrados en ii
	//y el numero de ellos en num_rsp
	num_rsp = hci_inquiry(dev_id, len, max_rsp, NULL, &ii, flags);
	//Comprobamos que la búsqueda se haya hecho con éxito
	if (num_rsp < 0) {
		//Liberar memoria usada en la busqueda de dispositivos
		free(ii);
		//Cerrar el socket con el microcontrolador
		hci_close_dev(sock);
		//Devolver puntero nulo como lista de dispositivos
		*list = NULL;
		//Devolver el error
		return E_SEARCH_DISP;
	}	

	//Recorremos los dispositivos encontrados
	for (i = 0; i < num_rsp; i++) {
		
		//Extrar la mac y puerto del dispositivo
		ba2str(&(ii[i].bdaddr), mac_aux);
		port_aux = get_obex_channel(mac_aux);
		
		//Si el dispositivo exporta el servicio FTP lo añadimos a la lista
		if (port_aux > 0) {
		
			//Aumentamos el numero de dispositivos que exportan FTP
			num_ftp++;
			
			//Reservamos memoria
			lista = (device_t *) realloc(lista, sizeof(device_t)*num_ftp);
			
			//Obtenemos el nombre corto
			if (hci_read_remote_name(sock, &(ii[i].bdaddr), 255, lista[num_ftp-1].name, 0) < 0)
				strcpy(lista[num_ftp-1].name, "[sin nombre]");
			
			//Asignamos mac y puerto obtenidos antes
			strcpy(lista[num_ftp-1].mac, mac_aux);
			lista[num_ftp-1].port = port_aux;
			
		}
		
	}

	//Liberar memoria usada en la busqueda de dispositivos
	free(ii);
	
	//Cerrar el socket con el microcontrolador
	hci_close_dev(sock);
	
	//Devolver lista de dispositivos
	*list = lista;
	
	return num_ftp;
}

int bftp_file_exist(char *remote_route) {
	if (obexftp_stat(cli, remote_route) == NULL)
		return E_FTP_EXIST;
	return 0;
}
